from typing import Mapping

from google.auth import jwt
import requests

from trailblazer.services.user_verification_service.exc import (
    GoogleCertsError,
    UserTokenVerificationError,
)
from trailblazer.store.models import User
from trailblazer.store.store import Store


class UserVerificationService:
    """Service to verify the user."""

    def __init__(self, store: Store, oauth_client_id: str):
        self.store: Store = store
        self.oauth_client_id: str = oauth_client_id

    def verify_user(self, authorization_header: str) -> User:
        """Verify the user by checking if the JWT token provided is valid."""
        jwt_token: str = self._extract_token_from_header(authorization_header)
        google_certs: Mapping = self._get_google_certs()
        try:
            payload: Mapping = jwt.decode(
                token=jwt_token,
                certs=google_certs,
                verify=True,
                audience=[self.oauth_client_id, "trailblazer"],
            )
        except Exception as error:
            raise UserTokenVerificationError(f"{error}") from error
        user_email: str = payload["email"]
        return self._get_user(user_email)

    @staticmethod
    def _extract_token_from_header(authorization_header: str) -> str:
        """Extract the token from the authorization header.
        args: authorization_header: The authorization header.
        raises ValueError: If no authorization header is provided.
        """
        if authorization_header:
            jwt_token = authorization_header.split("Bearer ")[-1]
            return jwt_token
        raise ValueError("No authorization header provided with request")

    @staticmethod
    def _get_google_certs() -> Mapping:
        """Get the Google certificates."""
        try:
            response = requests.get("https://www.googleapis.com/oauth2/v1/certs")
            response.raise_for_status()
            return response.json()
        except requests.RequestException as e:
            raise GoogleCertsError("Failed to fetch Google certs") from e

    def _get_user(self, user_email: str) -> User:
        """Check if the user is known."""
        if user := self.store.get_user(email=user_email, exclude_archived=True):
            return user
        raise ValueError("User not found in the database")



import subprocess

from trailblazer.constants import CharacterFormat, FileFormat
from trailblazer.exc import EmptySqueueError
from trailblazer.io.controller import ReadStream
from trailblazer.clients.slurm_cli_client.models import SqueueResult


def cancel_slurm_job(slurm_id: int, analysis_host: str | None = None) -> None:
    """Cancel SLURM job by SLURM job id."""
    scancel_commands: list[str] = ["scancel", str(slurm_id)]
    if analysis_host:
        scancel_commands = ["ssh", analysis_host] + scancel_commands
    subprocess.Popen(scancel_commands)


def get_slurm_queue(job_ids: list[int], analysis_host: str | None = None) -> SqueueResult:
    """Return squeue output from ongoing analyses in SLURM."""
    job_ids: str = ",".join(map(str, job_ids))
    queue_output: str = get_slurm_queue_output(job_ids=job_ids, analysis_host=analysis_host)
    return get_squeue_result(queue_output)


def get_slurm_queue_output(job_ids: str, analysis_host: str | None = None) -> str:
    """Return squeue output from ongoing analyses in SLURM."""
    squeue_commands: list[str] = [
        "squeue",
        "--jobs",
        job_ids,
        "--states=all",
        "--format",
        "%A,%j,%T,%l,%M,%S",
    ]
    if analysis_host:
        squeue_commands = ["ssh", analysis_host] + squeue_commands
        return (
            subprocess.check_output(
                squeue_commands,
                universal_newlines=True,
            )
            .decode(CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8)
            .strip()
            .replace("//n", "/n")
        )
    return subprocess.check_output(squeue_commands).decode(
        CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8
    )


def get_squeue_result(squeue_response: str) -> SqueueResult:
    """Return SqueueResult object from squeue response.
    Raises:
        TrailblazerError: when no entries were returned by squeue command.
    """
    if not squeue_response:
        raise EmptySqueueError("No jobs found in SLURM registry")
    squeue_response_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_response,
        read_to_dict=True,
    )
    return SqueueResult(jobs=squeue_response_content)



from trailblazer.clients.slurm_cli_client.utils import get_squeue_result
from trailblazer.clients.slurm_cli_client.models import SqueueResult


def test_get_squeue_result(squeue_stream_jobs):
    # GIVEN a squeue stream

    # WHEN getting the squeue results
    squeue_result: SqueueResult = get_squeue_result(squeue_response=squeue_stream_jobs)

    # THEN it should return squeue results and jobs
    assert isinstance(squeue_result, SqueueResult)
    assert isinstance(squeue_result.jobs, list)



import pytest

from trailblazer.clients.tower.models import TowerWorkflowResponse
from trailblazer.constants import TrailblazerStatus, WorkflowManager
from trailblazer.exceptions import NoJobsError
from trailblazer.services.job_service import JobService
from trailblazer.store.models import Analysis


def test_job_progression_completed(job_service: JobService, analysis_with_completed_jobs: Analysis):
    # GIVEN an analysis with completed jobs

    # WHEN getting the analysis progression
    progression = job_service.get_analysis_progression(analysis_with_completed_jobs.id)

    # THEN the progression is 100%
    assert progression == 1.0


def test_job_progression_ongoing(job_service: JobService, analysis_with_running_jobs: Analysis):
    # GIVEN an analysis with ongoing jobs

    # WHEN getting the analysis progression
    progression = job_service.get_analysis_progression(analysis_with_running_jobs.id)

    # THEN the progression is less than 100%
    assert progression < 1.0


def test_job_progression_without_jobs(job_service: JobService, analysis_without_jobs: Analysis):
    # GIVEN an analysis without jobs

    # WHEN getting the analysis progression
    progression = job_service.get_analysis_progression(analysis_without_jobs.id)

    # THEN the progression is 0%
    assert progression == 0.0


def test_analysis_status_when_completed_jobs(
    job_service: JobService,
    analysis_with_completed_jobs: Analysis,
):
    # GIVEN an analysis with completed jobs

    # WHEN getting the analysis status
    status = job_service.get_analysis_status(analysis_with_completed_jobs.id)

    # THEN the status is completed
    assert status == TrailblazerStatus.COMPLETED


def test_analysis_status_when_running_jobs(
    job_service: JobService,
    analysis_with_running_jobs: Analysis,
):
    # GIVEN an analysis with running jobs

    # WHEN getting the analysis status
    status = job_service.get_analysis_status(analysis_with_running_jobs.id)

    # THEN the status is running
    assert status == TrailblazerStatus.RUNNING


def test_fetch_pending_status_for_tower_without_jobs(
    job_service: JobService,
    tower_response_submitted: TowerWorkflowResponse,
):
    """
    Verify that a Tower-managed analysis with no associated jobs returns a PENDING status
    when Tower reports the workflow as submitted.
    """
    # GIVEN a Tower analysis with workflow manager nf_tower and without jobs
    analysis: Analysis = job_service.store.get_query(Analysis).first()
    analysis.workflow_manager = WorkflowManager.TOWER
    assert not analysis.jobs

    # GIVEN a simulated response from tower with the return status being submitted
    job_service.tower_service.client.get_workflow.return_value = tower_response_submitted

    # WHEN fetching the analysis status
    status: TrailblazerStatus = job_service.get_analysis_status(analysis.id)

    # THEN the analysis status should be PENDING
    assert status == TrailblazerStatus.PENDING


def test_no_jobs_error_for_slurm_analysis(job_service: JobService, analysis_without_jobs: Analysis):
    """
    Ensure that a NoJobsError is raised for a SLURM analysis with no associated jobs.
    """
    # GIVEN an analysis without any associated jobs
    # WHEN fetching the analysis status
    # THEN a NoJobsError should be raised
    with pytest.raises(NoJobsError):
        job_service.get_analysis_status(analysis_without_jobs.id)
